﻿using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;

new WindowManager();

internal class WindowManager
{
    internal Dictionary<IntPtr, WINDOWPLACEMENT> placements = new();
    internal Dictionary<IntPtr, string> windowTitles = new();
    internal Dictionary<IntPtr, Process> processes = new();
    private const int cutoffStr = 17;
    public WindowManager()
    {
        EnumWindows(WindowPlacement, 0);
        uint hwndCount = 1;
        //Group and print by pid
        foreach (int pid in processes.DistinctBy(kv => kv.Value.Id).Select(kv => kv.Value.Id).ToArray())
        {
            foreach(IntPtr hwnd in processes.Where(kv => kv.Value.Id == pid).Select(x => x.Key))
                Console.WriteLine($"{hwndCount++,3} {hwnd,8} {processes[hwnd].Id,8} {processes[hwnd].ProcessName.Substring(0, Math.Min(cutoffStr, processes[hwnd].ProcessName.Length)),cutoffStr} {windowTitles[hwnd].Substring(0, Math.Min(cutoffStr, windowTitles[hwnd].Length)),-cutoffStr} {placements[hwnd]}");
        }
    }

    private bool WindowPlacement(IntPtr hwnd, IntPtr lParam)
    {
        WINDOWPLACEMENT placement = new ();
        GetWindowPlacement(hwnd, ref placement);
        RECT n = placement.NormalPosition;
        //Do not add if Window is not a drawable
        if (n is { Left: 0, Top: 0, Right: 0, Bottom: 0 })
            return true;
        
        uint pid;
        GetWindowThreadProcessId(hwnd, out pid);
        Process process = Process.GetProcessById((int)pid);
        //nvcontainer does not have a window name
        if (process.ProcessName.Equals("nvcontainer"))
            return true;
        string name = GetWindowTextRaw(hwnd);
        //Do not add, if window does not have a name
        if (name.Length < 1)
            return true;
        
        processes.Add(hwnd, process);
        windowTitles.Add(hwnd, name);
        placements.Add(hwnd, placement);
        return true;
    }
    
    [DllImport("user32.dll")]
    [return: MarshalAs(UnmanagedType.Bool)]
    static extern bool EnumWindows(EnumWindowsProc lpEnumFunc, IntPtr lParam);

    private delegate bool EnumWindowsProc(IntPtr hWnd, IntPtr lParam);
    
    [DllImport("user32.dll")]
    [return: MarshalAs(UnmanagedType.Bool)]
    public static extern bool GetWindowPlacement(IntPtr hWnd, ref WINDOWPLACEMENT lpwndpl);
    
    [DllImport("user32.dll")]
    private static extern uint GetWindowThreadProcessId(IntPtr hWnd, out uint lpdwProcessId);

    
    [DllImport("user32.dll", CharSet = CharSet.Auto)]
    static extern IntPtr SendMessage(IntPtr hWnd, UInt32 Msg, IntPtr wParam, [Out] StringBuilder lParam);
    
    public static string GetWindowTextRaw(IntPtr hwnd)
    {
        UInt32 WM_GETTEXTLENGTH = 0x000E;
        UInt32 WM_GETTEXT = 0x000D;
        // Allocate correct string length first
        int length = (int)SendMessage(hwnd, WM_GETTEXTLENGTH, IntPtr.Zero, null);
        StringBuilder sb = new (length + 1);
        SendMessage(hwnd, WM_GETTEXT, sb.Capacity, sb);
        return sb.ToString();
    }
}

